"""Coordinax vectors.

Note that this module is private. Users should use the public API.

This module depends on the following modules:

- utils & typing
- angle
- distance


"""

__all__: list[str] = []
